/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      A sampler of the primitive addon.
 *      All primitives are rendered using the additive blender, so overdraw will manifest itself as overly bright pixels.
 *
 *
 *      By Pavel Sountsov.
 *
 *      See readme.txt for copyright information.
 */

#include <allegro5/allegro5.h>
#include <allegro5/a5_font.h>
#include "allegro5/a5_primitives.h"
#include <math.h>
#include <stdio.h>

typedef void (*Screen)(int, int, float);
int ScreenW = 800, ScreenH = 600;
#define NUM_SCREENS 5
#define ROTATE_SPEED 0.0001f
Screen Screens[NUM_SCREENS];
ALLEGRO_FONT* Font;
ALLEGRO_TRANSFORM Identity;
ALLEGRO_BITMAP* Buffer;

enum MODE {
   INIT,
   LOGIC,
   DRAW
};

float theta;
ALLEGRO_TRANSFORM main_trans;

void HighPrimitives(int mode, int soft, float speed)
{
   if (mode == INIT) {
   
   } else if (mode == LOGIC) {
      theta += speed;
      al_trans_build_2d(&main_trans, ScreenW / 2, ScreenH / 2, 1, 1, theta);
   } else if (mode == DRAW) {
      al_set_blender(ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA, al_map_rgba_f(1, 1, 1, 1));
      if (soft == 1)
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "High Level Primitives (Software)");
      else
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "High Level Primitives (Hardware)");
         
      al_set_blender(ALLEGRO_ONE, ALLEGRO_ONE, al_map_rgba_f(1, 1, 1, 1));
      
      if (soft == 1) {
         al_set_target_bitmap(Buffer);
         al_clear(al_map_rgba_f(0, 0, 0, 0));
      }
      
      al_trans_use_trans(&main_trans);
      
      al_draw_line_2d(-300, -200, 300, 200, al_map_rgba_f(0, 0.5, 0.5, 1), al_map_rgba_f(0, 0, 0, 1), 0);
      al_draw_triangle_2d(-150, -250, 0, 250, 150, -250, al_map_rgba_f(0.5, 0, 0.5, 1), 0, ALLEGRO_OUTLINED);
      al_draw_rectangle_2d(-300, -200, 300, 200, al_map_rgba_f(0.5, 0, 0, 0.5), 0, ALLEGRO_OUTLINED);
      al_draw_ellipse_2d(0, 0, 300, 150, al_map_rgba_f(0, 0, 0, 0.5), al_map_rgba_f(0, 0.5, 0.5, 1), 0, ALLEGRO_OUTLINED);
      al_draw_arc_2d(0, 0, 200, -AL_PI / 2, AL_PI, al_map_rgba_f(0.5, 0.25, 0, 1), al_map_rgba_f(0, 0.25, 0.5, 1), 0, ALLEGRO_OUTLINED);
      float points[8] = {
         -300, -100,
         700, -100,
         -700, 100,
         300, 100
      };
      al_draw_spline_2d(points, al_map_rgba_f(0.5, 0, 0, 1), al_map_rgba_f(0, 0.5, 0, 1), 0);
      
      al_trans_use_trans(&Identity);
      
      if (soft == 1) {
         al_set_target_bitmap(al_get_backbuffer());
         al_draw_bitmap(Buffer, 0, 0, 0);
      }
   }
}

void ShadePrimitives(int mode, int soft, float speed)
{
   static ALLEGRO_COLOR shade_color;
   float t = al_current_time();
   if (mode == INIT) {
   
   } else if (mode == LOGIC) {
      theta += speed;
      al_trans_build_2d(&main_trans, ScreenW / 2, ScreenH / 2, 1, 1, theta);
      shade_color = al_map_rgba_f(1 + 0.5 * sinf(t), 1 + 0.5 * sinf(t + AL_PI / 3), 1 + 0.5 * sinf(t + 2 * AL_PI / 3), 1);
   } else if (mode == DRAW) {
      al_set_blender(ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA, shade_color);
      if (soft == 1)
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Shaded Primitives (Software)");
      else
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Shaded Primitives (Hardware)");
         
      al_set_blender(ALLEGRO_ONE, ALLEGRO_ONE, shade_color);
      
      if (soft == 1) {
         al_set_target_bitmap(Buffer);
         al_clear(al_map_rgba_f(0, 0, 0, 0));
      }
      
      al_trans_use_trans(&main_trans);
      
      al_draw_line_2d(-300, -200, 300, 200, al_map_rgba_f(0, 0.5, 0.5, 1), al_map_rgba_f(0, 0, 0, 1), 0);
      al_draw_triangle_2d(-150, -250, 0, 250, 150, -250, al_map_rgba_f(0.5, 0, 0.5, 1), 0, ALLEGRO_OUTLINED);
      al_draw_rectangle_2d(-300, -200, 300, 200, al_map_rgba_f(0.5, 0, 0, 0.5), 0, ALLEGRO_OUTLINED);
      al_draw_ellipse_2d(0, 0, 300, 150, al_map_rgba_f(0, 0, 0, 0.5), al_map_rgba_f(0, 0.5, 0.5, 1), 0, ALLEGRO_OUTLINED);
      al_draw_arc_2d(0, 0, 200, -AL_PI / 2, AL_PI, al_map_rgba_f(0.5, 0.25, 0, 1), al_map_rgba_f(0, 0.25, 0.5, 1), 0, ALLEGRO_OUTLINED);
      float points[8] = {
         -300, -100,
         700, -100,
         -700, 100,
         300, 100
      };
      al_draw_spline_2d(points, al_map_rgba_f(0.5, 0, 0, 1), al_map_rgba_f(0, 0.5, 0, 1), 0);
      
      al_trans_use_trans(&Identity);
      
      if (soft == 1) {
         al_set_target_bitmap(al_get_backbuffer());
         al_draw_bitmap(Buffer, 0, 0, 0);
      }
   }
}

void TransformationsPrimitives(int mode, int soft, float speed)
{
   float t = al_current_time();
   if (mode == INIT) {
   
   } else if (mode == LOGIC) {
      theta += speed;
      al_trans_build_2d(&main_trans, ScreenW / 2, ScreenH / 2, sinf(t / 5), cosf(t / 5), theta);
   } else if (mode == DRAW) {
      al_set_blender(ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA, al_map_rgba_f(1, 1, 1, 1));
      if (soft == 1)
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Transformations (Software)");
      else
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Transformations (Hardware)");
         
      al_set_blender(ALLEGRO_ONE, ALLEGRO_ONE, al_map_rgba_f(1, 1, 1, 1));
      
      if (soft == 1) {
         al_set_target_bitmap(Buffer);
         al_clear(al_map_rgba_f(0, 0, 0, 0));
      }
      
      al_trans_use_trans(&main_trans);
      
      al_draw_line_2d(-300, -200, 300, 200, al_map_rgba_f(0, 0.5, 0.5, 1), al_map_rgba_f(0, 0, 0, 1), 0);
      al_draw_triangle_2d(-150, -250, 0, 250, 150, -250, al_map_rgba_f(0.5, 0, 0.5, 1), 0, ALLEGRO_OUTLINED);
      al_draw_rectangle_2d(-300, -200, 300, 200, al_map_rgba_f(0.5, 0, 0, 0.5), 0, ALLEGRO_OUTLINED);
      al_draw_ellipse_2d(0, 0, 300, 150, al_map_rgba_f(0, 0, 0, 0.5), al_map_rgba_f(0, 0.5, 0.5, 1), 0, ALLEGRO_OUTLINED);
      al_draw_arc_2d(0, 0, 200, -AL_PI / 2, AL_PI, al_map_rgba_f(0.5, 0.25, 0, 1), al_map_rgba_f(0, 0.25, 0.5, 1), 0, ALLEGRO_OUTLINED);
      float points[8] = {
         -300, -100,
         700, -100,
         -700, 100,
         300, 100
      };
      al_draw_spline_2d(points, al_map_rgba_f(0.5, 0, 0, 1), al_map_rgba_f(0, 0.5, 0, 1), 0);
      
      al_trans_use_trans(&Identity);
      
      if (soft == 1) {
         al_set_target_bitmap(al_get_backbuffer());
         al_draw_bitmap(Buffer, 0, 0, 0);
      }
   }
}


void LowPrimitives(int mode, int soft, float speed)
{
   static ALLEGRO_VBUFFER* vbuff;
   if (mode == INIT) {
      vbuff = al_create_vbuff(13, ALLEGRO_VBUFFER_SOFT | ALLEGRO_VBUFFER_WRITE | ALLEGRO_VBUFFER_READ);
      int ii = 0;
      for (ii = 0; ii < 13; ii++) {
         float x, y;
         x = 200 * cosf((float)ii / 13.0f * 2 * AL_PI);
         y = 200 * sinf((float)ii / 13.0f * 2 * AL_PI);
         
         ALLEGRO_COLOR color = al_map_rgb((ii + 1) % 3 * 64, (ii + 2) % 3 * 64, (ii) % 3 * 64);
         
         al_prim_set_pos(vbuff, ii, x, y, 0);
         al_prim_set_color(vbuff, ii, color);
      }
   } else if (mode == LOGIC) {
      theta += speed;
      al_trans_build_2d(&main_trans, ScreenW / 2, ScreenH / 2, 1, 1, theta);
   } else if (mode == DRAW) {
      al_set_blender(ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA, al_map_rgba_f(1, 1, 1, 1));
      if (soft == 1)
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Low Level Primitives (Software)");
      else
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Low Level Primitives (Hardware)");
         
      al_set_blender(ALLEGRO_ONE, ALLEGRO_ONE, al_map_rgba_f(1, 1, 1, 1));
      
      if (soft == 1) {
         al_set_target_bitmap(Buffer);
         al_clear(al_map_rgba_f(0, 0, 0, 0));
      }
      
      al_trans_use_trans(&main_trans);
      
      al_prim_draw(vbuff, 0, 0, 4, ALLEGRO_PRIM_LINE_LIST);
      
      al_prim_draw(vbuff, 0, 4, 9, ALLEGRO_PRIM_LINE_STRIP);
      
      al_prim_draw(vbuff, 0, 9, 13, ALLEGRO_PRIM_LINE_LOOP);
      
      al_trans_use_trans(&Identity);
      
      if (soft == 1) {
         al_set_target_bitmap(al_get_backbuffer());
         al_draw_bitmap(Buffer, 0, 0, 0);
      }
   }
}

void IndexedPrimitives(int mode, int soft, float speed)
{
   static ALLEGRO_VBUFFER* vbuff;
   static int indices[4] = {0, 1, 3, 4};
   if (mode == INIT) {
      vbuff = al_create_vbuff(13, ALLEGRO_VBUFFER_SOFT | ALLEGRO_VBUFFER_WRITE | ALLEGRO_VBUFFER_READ);
      int ii = 0;
      for (ii = 0; ii < 13; ii++) {
         float x, y;
         x = 200 * cosf((float)ii / 13.0f * 2 * AL_PI);
         y = 200 * sinf((float)ii / 13.0f * 2 * AL_PI);
         
         ALLEGRO_COLOR color = al_map_rgb((ii + 1) % 3 * 64, (ii + 2) % 3 * 64, (ii) % 3 * 64);
         
         al_prim_set_pos(vbuff, ii, x, y, 0);
         al_prim_set_color(vbuff, ii, color);
      }
   } else if (mode == LOGIC) {
      int ii;
      theta += speed;
      for (ii = 0; ii < 4; ii++)
         indices[ii] = ((int)al_current_time() + ii) % 13;
         
      al_trans_build_2d(&main_trans, ScreenW / 2, ScreenH / 2, 1, 1, theta);
   } else if (mode == DRAW) {
      int ii;
      al_set_blender(ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA, al_map_rgba_f(1, 1, 1, 1));
      if (soft == 1)
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Indexed Primitives (Software)");
      else
         al_font_textprintf_centre(Font, ScreenW / 2, 0, "Indexed Primitives (Hardware)");
         
      al_set_blender(ALLEGRO_ONE, ALLEGRO_ONE, al_map_rgba_f(1, 1, 1, 1));
      
      if (soft == 1) {
         al_set_target_bitmap(Buffer);
         al_clear(al_map_rgba_f(0, 0, 0, 0));
      }
      
      al_trans_use_trans(&main_trans);
      
      al_prim_draw_indexed(vbuff, 0, indices, 4, ALLEGRO_PRIM_LINE_LIST);
      
      for (ii = 0; ii < 4; ii++) {
         indices[ii] += 4;
         indices[ii] %= 13;
      }
      al_prim_draw_indexed(vbuff, 0, indices, 4, ALLEGRO_PRIM_LINE_STRIP);
      
      for (ii = 0; ii < 4; ii++) {
         indices[ii] += 4;
         indices[ii] %= 13;
      }
      al_prim_draw_indexed(vbuff, 0, indices, 4, ALLEGRO_PRIM_LINE_LOOP);
      
      al_trans_use_trans(&Identity);
      
      if (soft == 1) {
         al_set_target_bitmap(al_get_backbuffer());
         al_draw_bitmap(Buffer, 0, 0, 0);
      }
   }
}

int main(int argc, char *argv[])
{
   // Initialize Allegro 5 and the font routines
   al_init();
   al_font_init();
   
   // Create a window to display things on: 640x480 pixels
   ALLEGRO_DISPLAY *display = al_create_display(800, 600);
   if (!display) {
      printf("Error creating display.\n");
      return 1;
   }
   
   // Install the keyboard handler
   if (!al_install_keyboard()) {
      printf("Error installing keyboard.\n");
      return 1;
   }
   
   // Load a font
   Font = al_ttf_load_font("data/DejaVuSans.ttf", 16, 0);
   if (!Font) {
      printf("Error loading \"data/DejaVuSans.ttf\".\n");
      return 1;
   }
   
   // Make and set some color to draw with
   ALLEGRO_COLOR white = al_map_rgba_f(1.0, 1.0, 1.0, 1.0);
   ALLEGRO_COLOR black = al_map_rgba_f(0.0, 0.0, 0.0, 1.0);
   
   // Start the event queue to handle keyboard input and our timer
   ALLEGRO_EVENT_QUEUE *queue = al_create_event_queue();
   al_register_event_source(queue, (ALLEGRO_EVENT_SOURCE*)al_get_keyboard());
   
   int refresh_rate = 60;
   int frames_done = 0;
   double time_diff = al_current_time();
   double fixed_timestep = 1.0f / refresh_rate;
   double real_time = al_current_time();
   double game_time = al_current_time();
   int ii;
   
   ALLEGRO_TIMER *timer = al_install_timer(ALLEGRO_BPS_TO_SECS(refresh_rate));
   al_start_timer(timer);
   ALLEGRO_EVENT_QUEUE *timer_queue = al_create_event_queue();
   al_register_event_source(timer_queue, (ALLEGRO_EVENT_SOURCE*)timer);
   
   int old = al_get_new_bitmap_flags();
   al_set_new_bitmap_flags(ALLEGRO_MEMORY_BITMAP);
   Buffer = al_create_bitmap(ScreenW, ScreenH);
   al_set_new_bitmap_flags(old);
   
   al_trans_set_identity(&Identity);
   int cur_screen = 0;
   int use_software = 0;
   float speed = ROTATE_SPEED;
   
   Screens[0] = LowPrimitives;
   Screens[1] = IndexedPrimitives;
   Screens[2] = HighPrimitives;
   Screens[3] = ShadePrimitives;
   Screens[4] = TransformationsPrimitives;
   
   for (ii = 0; ii < NUM_SCREENS; ii++)
      Screens[ii](INIT, 0, speed);
      
   bool done = false;
   while (!done) {
      double frame_duration = al_current_time() - real_time;
      al_rest(fixed_timestep - frame_duration); //rest at least fixed_dt
      frame_duration = al_current_time() - real_time;
      real_time = al_current_time();
      
      if (real_time - game_time > frame_duration) { //eliminate excess overflow
         game_time += fixed_timestep * floor((real_time - game_time) / fixed_timestep);
      }
      
      while (real_time - game_time >= 0) {
         double start_time = al_current_time();
         game_time += fixed_timestep;
         
         Screens[cur_screen](LOGIC, use_software, speed);
         
         ALLEGRO_EVENT key_event;
         while (al_get_next_event(queue, &key_event)) {
            switch (key_event.type) {
            case ALLEGRO_EVENT_KEY_DOWN: {
                  switch (key_event.keyboard.keycode) {
                  case ALLEGRO_KEY_ESCAPE: {
                        done = true;
                        break;
                     }
                  case ALLEGRO_KEY_S: {
                        use_software = !use_software;
                        time_diff = al_current_time();
                        frames_done = 0;
                        break;
                     }
                  case ALLEGRO_KEY_LEFT: {
                        speed -= ROTATE_SPEED;
                        break;
                     }
                  case ALLEGRO_KEY_RIGHT: {
                        speed += ROTATE_SPEED;
                        break;
                     }
                  case ALLEGRO_KEY_UP:
                  case ALLEGRO_KEY_DOWN: {
                        speed = 0;
                        break;
                     }
                  default: {
                        cur_screen++;
                        if (cur_screen >= NUM_SCREENS) {
                           cur_screen = 0;
                        }
                     }
                  }
               }
            }
         }
         
         if (al_current_time() - start_time >= fixed_timestep) { //break if we start taking too long
            break;
         }
      }
      al_clear(black);
      al_set_blender(ALLEGRO_ALPHA, ALLEGRO_INVERSE_ALPHA, white);
      al_font_textprintf(Font, 0, 0, "FPS: %f", (float)frames_done / (al_current_time() - time_diff));
      al_font_textprintf(Font, 0, 20, "Try arrows, S, Space. Esc to Quit.");
      
      Screens[cur_screen](DRAW, use_software, speed);
      
      al_wait_for_vsync();
      al_flip_display();
      frames_done++;
   }
   
   return 0;
}
END_OF_MAIN()
